# Generated by Django 4.0.dev20210811195242 on 2021-08-13 10:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        (
            "migration_test_app",
            "0003_alter_historicalmodelwithcustomattrforeignkey_options_and_more",
        ),
    ]

    operations = [
        migrations.AlterField(
            model_name="historicalmodelwithcustomattrforeignkey",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalyar",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]
