{% extends "admin/object_history.html" %}
{% load i18n %}
{% load url from simple_history_compat %}
{% load admin_urls %}
{% load display_list from simple_history_admin_list %}


{% block content %}
  <div id="content-main">
    {% if not revert_disabled %}<p>
      {% blocktrans %}Choose a date from the list below to revert to a previous version of this object.{% endblocktrans %}</p>{% endif %}
    <div class="module">
      {% if action_list %}
        {% display_list %}
      {% else %}
        <p>{% trans "This object doesn't have a change history." %}</p>
      {% endif %}
    </div>
  </div>
{% endblock %}
